
#include "stdafx.h"

#include <dxmus.core/Controller.h>
#include "DmPortPlayer.h"



namespace GuruGuruSmf { namespace Core { namespace DirectMusicPort {

	// XbhĂяop
	static DWORD WINAPI PlayerWorker(LPVOID object)
	{
		((Player*)object)->OnWorker();
		return 0;
	}

	// RXgN^
	Player::Player()
	{
		times = new PlayerTimes();
		device = new Device(times);
		volume = new Volume(device);

		workerThread = new Thread();

		sentReset = false;
		isWorkerOn = false;
		doMasterPitch = false;
		loopCount = 0;
		state = PlayerState::Stop;
	}

	// fXgN^
	Player::~Player()
	{
		TRACE(L"DmPort.Player:fXgN^:CloseDevice\n");
		CloseDevice();

		TRACE(L": volume\n");
		delete volume;
		TRACE(L": device\n");
		delete device;
		TRACE(L": times\n");
		delete times;

		TRACE(L": thread\n");
		delete workerThread;
	}

	// foCXJ
	GGSERROR Player::OpenDevice(int deviceId, HWND handle)
	{
		CloseDevice();

		GGS_LOGGING_INITIALIZE();

		return device->Open(deviceId, handle);
	}

	// foCX
	void Player::CloseDevice()
	{
		if (!device->GetIsOpen()) return;
		
		TRACE(L"DmPort.Player: CloseDevice : Stop\n");
		Stop(0);
		TRACE(L"DmPort.Player: CloseDevice : Sleep\n");
		Sleep(150);
		TRACE(L"DmPort.Player: CloseDevice : device->Close()\n");
		device->Close();
	}

	// tI
	GGSERROR Player::Play(ISmfController* smfCtrl, int option, int seek, int fadeInTime, int newKeyShift)
	{
		if (!device->GetIsOpen()) return GgsError::NotReady;
		Stop(0);

		Smf* newSmf = smfCtrl->GetSmf();

		if (!newSmf->GetIsReady()) return GgsError::NotReady;
		if (newSmf->GetData()->FirstNoteTick == -1) return GgsError::NoError;	// f[^̓X[

		GGSERROR err = smfCtrl->Open();
		if(err != GgsError::NoError) return err;
		smfController = smfCtrl;

		Initialize(newSmf);

		VoiceList* voiceList = newSmf->GetData()->VoiceList;
		GuruGuruSmf::Dxmus::Controller::GetInstance()->PortSetVoiceList((DWORD*)voiceList->GetItems(), voiceList->GetCount());
		ParsePlayOption(option);
		SetKeyShift(newKeyShift);
		
		
		// OɉMĂ1~b̃EFCg
		times->ReadMasterClock();
		times->SetMarkerDmTimeByQueueDmTime(10000);
	
		// StopȂǂŃZbg𑗂ĂȂꍇAGSZbg𑗂Ă
		if (!sentReset){
			device->SendLongMessage(Exclusive::GSReset, Exclusive::GSResetLength);
			times->Wait(10000);		// svȂ悤łꉞ1~bEFCg
		}
		sentReset = false;
		
		// RPN0Ԃɂ̓ZbgȂ悤Ȃ̂őMĂ
		InitializePitchBendSensitivity(device);
		
		volume->SetFadeIn(fadeInTime);
		volume->InitializeModule(GetDoNormalize(), newSmf->GetData()->MaxVolume);
		times->Wait(10000);		// Ƃ悤Ȃ̂ŃEFCg
		
		if (seek > 0){
			Seek(seek);
		}else{
			SkipBeginning();
		}

		times->UpdateQueueDmTime();
	
		if (TimeEventStep()) return GgsError::NoError;
		state = PlayerState::Play;
		StartWorker();
		
		return GgsError::NoError;
	}

	// ~
	void Player::Stop(int fadeOutTime)
	{
		if (!device->GetIsOpen() || state == PlayerState::Stop) return;
		
		// ꎞ~Ԃ̎͂ɒ~
		if (state != PlayerState::Pause && fadeOutTime > 0){	// tF[hAEg
			volume->SetFadeOut(fadeOutTime);
			return;
		}
		
		StopWorker();
		FinalizeMusic(false, true);
	}

	// ꎞ~
	void Player::Pause()
	{
		if (state == PlayerState::Play){
			isPause = true;
			doRestart = false;
			Sleep(Constant::TimeInterval);	// ی
		}
	}

	// ꎞ~
	void Player::Restart()
	{
		if (state == PlayerState::Pause){
			doRestart = true;
			Sleep(Constant::TimeInterval);	// ی
		}
	}

	// t擾
	void Player::GetStatus(PlayerStatus* status)
	{
		status->State = state;
		status->LoopCount = loopCount;
		status->Tick = times->GetNowTick();
		status->Time = (int)times->GetNowTime();
		status->Tempo = times->GetMusicTempo();
	}

	// }X^[e|
	void Player::SetMasterTempo(float masterTempo)
	{
		times->SetMasterTempo(masterTempo);
	}

	// }X^[{[
	void Player::SetMasterVolume(int masterVolume)
	{
		volume->SetMasterVolume(masterVolume);
	}

	// }X^[sb`
	void Player::SetMasterPitch(int pitch)
	{
		SetMasterPitchBase(pitch);
		doMasterPitch = true;
	}
		
	// }[J[TickɉZ
	void Player::AddMarkerTick(int deltaTime)
	{
		times->UpdateMarkerTick(deltaTime);
		
		// [vʒuݒ
		if (loopPosition == -1 && times->PassedLoopTick()){
			loopPosition = GetCode()->GetPosition();
		}
	}
		
	// I_
	void Player::MusicEnd()
	{
		isMusicEnd = true;
	}

	// e|Zbg
	void Player::SetTempo(int tempo)
	{
		times->SetTempo(tempo);
	}

	// {[bZ[W𑗐M
	void Player::SendVolume(int channel, int vol)
	{
		volume->SetVolume(channel, vol);
	}
		
	// sb`xhbZ[W𑗐M
	void Player::SendPitchBend(int channel, int pitch)
	{
		SendPitchBendBase(device, channel, pitch);
	}

	// GNXN[Vu𑗐M
	void Player::SendExclusive(CodeEvent* codeEvent)
	{
		SendExclusiveBase(device, codeEvent);
	}
		
	// SeekpɃCxgo^
	void Player::RegisterEvent(CODE_EVENTTYPE type, int channel, int data1, CodeEvent* codeEvent)
	{
		int key = (byte)type << 16 | channel << 8 | data1;
		seekingCodeEventTable[key] = *codeEvent;
	}

	// Vtgꂽm[gԍ擾
	int Player::GetShiftedNote(int note, int channel)
	{
		return GetShiftedNoteBase(note, channel);
	}
		
	// ݎ+PrepareTime܂ł̃Cxg
	bool Player::TimeEventStep()
	{
		while(!times->PassedPrepareTime()){
			CodeEvent ev;
			GetCode()->ReadEvent(&ev);
			PlayingEventExecute(&ev, this);
			
			// ̋󔒔΂
			if(GetDoSkipEnding()){
				if(times->PassedLastNote()){
					times->RewindQueueTime();	// ςݎԂ܂Ŗ߂Ă
					isMusicEnd = true;
				}
			}
			
			// [vI
			if(isMusicEnd){
				++loopCount;
				if(GetDoLoop()){
					switch(GetSmf()->GetData()->LoopType){
					case LoopType::Beginning:
						Rewind();
						SkipBeginning();
						break;
					case LoopType::LoopPointAtBeginning:
						Rewind();
						break;
					case LoopType::LoopPoint:
						GetCode()->SetPosition(loopPosition);
						times->GotoLoopPoint();
						isMusicEnd = false;
						break;
					case LoopType::End:
						return true;
					}
				}else{
					return true;	// RI
				}
			}
		}
		
		// ݂TickƉtԂvZ
		times->CalcNowTime();
		return false;
	}
		
	// `̋󔒂XLbv
	void Player::SkipBeginning()
	{
		if(GetDoSkipBeginning()){
			times->SetSkipping(true);
			
			while(!times->PassedFirstNote()){
				CodeEvent ev;
				GetCode()->ReadEvent(&ev);
				PlayingEventExecute(&ev, this);
			}
			times->SetSkipping(false);
		}
	}
		
	// V[N
	void Player::Seek(int tick)
	{
		seekingCodeEventTable.clear();
		times->SetSkipping(true);
		while(!times->PassedTick(tick)){
			CodeEvent ev;
			GetCode()->ReadEvent(&ev);
			SeekingEventExecute(&ev, this);
			if (isMusicEnd) break;
		}
		
		// ۑĂf[^𑗐M
		std::map<int, CodeEvent>::iterator it = seekingCodeEventTable.begin();
		while(it != seekingCodeEventTable.end()){
			PlayingEventExecute(&(*it).second, this);
			++it;
		}
		seekingCodeEventTable.clear();

		times->SetSkipping(false);
	}
	
	// XbhJn
	void Player::StartWorker()
	{
		workerThread->Join();	// XbhSɏIĂȂP[X̂ő҂

		isWorkerOn = true;
		workerThread->Start(this, PlayerWorker, THREAD_PRIORITY_TIME_CRITICAL);
	}

	// Xbh~
	void Player::StopWorker()
	{
		if (isWorkerOn){
			isWorkerOn = false;
			workerThread->Join();
		}
	}
		
	// ꎞ~
	void Player::WorkerPause()
	{
		if (state == PlayerState::Play){
			// ʂ0ɂ
			times->EnterRealtimeMode();
			volume->Mute(true);
			times->LeaveRealtimeMode();
			times->Pause();
			
			state = PlayerState::Pause;
		}
	}
		
	// ꎞ~
	void Player::WorkerRestart()
	{
		times->Restart();
		
		times->EnterRealtimeMode();
		volume->Mute(false);
		times->LeaveRealtimeMode();
		
		state = PlayerState::Play;
		isPause = false;
		doRestart = false;
	}
		
	// Xbhs
	void Player::OnWorker()
	{
		while(isWorkerOn){
			times->ReadMasterClock();
			
			if (doRestart) WorkerRestart();
			
			if (isPause){
				WorkerPause();
			}else{
				times->EnterRealtimeMode();
				bool terminated = volume->Worker();
				times->LeaveRealtimeMode();
				if (terminated){
					// I[m[gIt
					for (int i=0; i<16; i++){
						device->SendShortMessage(0xb0+i, 123, 0);
					}
					times->UpdateQueueDmTime();
					FinalizeMusic(true, true);
					return;
				}
				
				if (GetCanDoMasterPitch() && doMasterPitch){
					SendMasterPitch();
					doMasterPitch = false;
				}
				
				if (TimeEventStep()){
					FinalizeMusic(true, false);
					return;
				}
			}
			
			Sleep(Constant::TimeInterval);
			times->FixQueueDmTime();
		}
	}
	
	// y~̍ŏI
	void Player::FinalizeMusic(bool isWorker, bool doReset)
	{
		if (isWorker){
			isWorkerOn = false;
		}
		state = PlayerState::Stop;
		smfController->Close();
	
		if (doReset){
			// GSZbg𑗂
			times->ReadMasterClock();
			times->SetMarkerDmTimeByQueueDmTime(10000);
			device->SendLongMessage(Exclusive::GSReset, Exclusive::GSResetLength);
			sentReset = true;
			times->UpdateQueueDmTime();
		}
	}
	
	// }X^[sb`𑗐M
	void Player::SendMasterPitch()
	{
		times->EnterRealtimeMode();
		SendMasterPitchBase(device);
		times->LeaveRealtimeMode();
	}
		
	// SMFŏ
	void Player::Initialize(Smf* newSmf)
	{
		PlayerBase::Initialize(newSmf);
		
		times->Initialize(newSmf);
		loopPosition = -1;
	
		loopCount = 0;
		isPause = false;
		doRestart = false;
	
		Rewind();
	}

	// `Ɋ߂
	void Player::Rewind()
	{
		PlayerBase::Rewind();
		isMusicEnd = false;

		times->Rewind();
		GetCode()->SetPosition(0);
	}



}}}

