// DirectMusic Port[hvC[

#pragma once

#include <Windows.h>
#include <map>

#include "Interface.h"
#include "Device.h"
#include "Volume.h"
#include "Times.h"
#include "Thread.h"
#include "Player.h"


namespace GuruGuruSmf { namespace Core { namespace DirectMusicPort {


	/// <summary>
	/// DirectMusic Port[h̃vC[
	/// </summary>
	class Player : public Core::Player::PlayerBase, public IPlayer, public IPlayerCore
	{
	private:
		Device* device;
		bool sentReset;
		
		Thread* workerThread;
		volatile bool isWorkerOn;
		
		ISmfController* smfController;

		std::map<int, CodeEvent> seekingCodeEventTable;
		bool doMasterPitch;
		
		PlayerTimes* times;
		Volume* volume;
		
		PLAYERSTATE state;
		int loopPosition, loopCount;
		bool isMusicEnd;
		
		bool isPause, doRestart;

		// ݎ+PrepareTime܂ł̃Cxg
		bool TimeEventStep();
		
		// `̋󔒂XLbv
		void SkipBeginning();
		
		// V[N
		void Seek(int tick);
		
		// XbhJn
		void StartWorker();

		// Xbh~
		void StopWorker();
		
		// ꎞ~
		void WorkerPause();
		
		// ꎞ~
		void WorkerRestart();
		
		// y~̍ŏI
		void FinalizeMusic(bool isWorker, bool doReset);
		
		// }X^[sb`𑗐M
		void SendMasterPitch();

	protected:
		// SMFŏ
		void Initialize(Smf* newSmf);

		// `Ɋ߂
		void Rewind();

	public:
		Player();	// RXgN^
		~Player();	// fXgN^

		// Xbhs
		void OnWorker();
		
		//
		// IPlayer
		//
		
		// foCXJ
		GGSERROR OpenDevice(int deviceId, HWND handle);
		
		// foCX
		void CloseDevice();
		
		// tI
		GGSERROR Play(ISmfController* smfCtrl, int option, int seek, int fadeInTime, int newKeyShift);
		
		// ~
		void Stop(int fadeOutTime);
		
		// ꎞ~
		void Pause();
		
		// ꎞ~
		void Restart();

		// t擾
		void GetStatus(PlayerStatus* status);
		
		// }X^[e|
		void SetMasterTempo(float masterTempo);
		
		// }X^[{[
		void SetMasterVolume(int masterVolume);
		
		// }X^[sb`
		void SetMasterPitch(int pitch);
		
		
		//
		// IPlayerCore
		//
		
		// MIDIbZ[WMfoCX擾
		IDevice* GetDevice(){ return device; }
		
		// }[J[TickɉZ
		void AddMarkerTick(int deltaTime);
		
		// I_
		void MusicEnd();
		
		// e|Zbg
		void SetTempo(int tempo);
		
		// {[bZ[W𑗐M
		void SendVolume(int channel, int vol);
		
		// sb`xhbZ[W𑗐M
		void SendPitchBend(int channel, int pitch);
		
		// GNXN[Vu𑗐M
		void SendExclusive(CodeEvent* codeEvent);
		
		// SeekpɃCxgo^
		void RegisterEvent(CODE_EVENTTYPE type, int channel, int data1, CodeEvent* codeEvent);

		// Vtgꂽm[gԍ擾
		int GetShiftedNote(int note, int channel);


	};

}}}

