
#include "stdafx.h"
#include "Times.h"
#include <dxmus.core/Controller.h>


namespace GuruGuruSmf { namespace Core {

namespace Analyzer {
	
	//
	// PlayerTimes
	//
	
	// `Ɋ߂
	void AnalyzerTimes::Rewind()
	{
		SetTempo(Constant::DefaultTempo);
		markerTick = 0;
		markerTime = 0;
	}
	
	// TickTimeXV
	void AnalyzerTimes::Update(int tick)
	{
		markerTime += (tick - markerTick) * tempoA;
		markerTick = tick;
	}

}

namespace WindowsMultimedia
{
	//
	// PlayerTimes
	//
	
	// RXgN^
	PlayerTimes::PlayerTimes()
	{
		masterTempo = 1;

		timeBase = 96;
		SetTempo(Constant::DefaultTempo);
		nowTime = 0;
		nowTick = 0;
	}
	
	// e|(SMFł̕\l)
	void PlayerTimes::SetTempo(double tp)
	{
		tempo = tp;
		tempoA = tp / (timeBase * 1000);
		deltaTick = 1000 * timeBase * Constant::TimeInterval / tp;
	}
	
	// MarkerTickXV
	void PlayerTimes::UpdateMarkerTick(int delta)
	{
		queueTick = markerTick;
		queueTime = markerTime;
		markerTick += delta;
		markerTime += delta * tempoA;
	}
	
	// ݂Ticki߂
	void PlayerTimes::UpdateNowTick()
	{
		nowTime += Constant::TimeInterval * masterTempo;
		nowTick += deltaTick * masterTempo;
	}
	
	// ݂̉tԂC V[NɎgp
	void PlayerTimes::FixNowTime(int tick)
	{
		nowTime = queueTime + (tick - queueTick) * tempoA;
		nowTick = tick;
	}
	
	// ŏ̔̈ʒuֈړ
	void PlayerTimes::GotoFirstNote()
	{
		nowTime = firstNoteTime;
		markerTime = nowTime;
		nowTick = firstNoteTick;
	}
	
	// [vʒuֈړ
	void PlayerTimes::GotoLoopPoint(bool isSkipEnding)
	{
		markerTick = loopTick;
		markerTime = loopTime;
		
		double sTime = totalTime;
		double sTick = totalTick;
		
		if (isSkipEnding){
			sTime = lastNoteTime;
			sTick = lastNoteTick;
		}
		nowTime = loopTime + (nowTime - sTime);
		nowTick = loopTick + (nowTick - sTick);
	}
	
	// SMFŏ
	void PlayerTimes::Initialize(Smf* smf)
	{
		timeBase = smf->GetTimeBase();
		
		loopTick = smf->GetData()->LoopTick;
		loopTime = smf->GetData()->LoopTime;
		firstNoteTick = smf->GetData()->FirstNoteTick;
		firstNoteTime = smf->GetData()->FirstNoteTime;
		lastNoteTick = smf->GetData()->LastNoteTick;
		lastNoteTime = smf->GetData()->LastNoteTime;
		totalTick = smf->GetData()->TotalTick;
		totalTime = smf->GetData()->TotalTime;
		
		queueTick = 0;
		queueTime = 0;
	}
	
	// `Ɋ߂
	void PlayerTimes::Rewind()
	{
		SetTempo(Constant::DefaultTempo);
		markerTick = 0;
		markerTime = 0;
		
		nowTick = 0;
		nowTime = 0;
	}
	
	
}

namespace DirectMusicPort {

	//
	// PlayerTimes
	//
	
	// RXgN^
	PlayerTimes::PlayerTimes()
	{
		queueDmTime = 0;
		fixedQueueDmTime = 0;
		masterTempo = 1;
		realtimeMode = false;

		timeBase = 96;
		SetTempo(Constant::DefaultTempo);
		nowTime = 0;
		nowTick = 0;
	}
	
	// ̃bZ[WM\莞
	__int64 PlayerTimes::GetMarkerDmTime(){
		if (realtimeMode){
			if(fixedQueueDmTime > markerDmTime-10000){
				// ʏ̏ԂǂȂ悤ɍ׍H
				markerDmTime = fixedQueueDmTime + 10000;
			}
			return fixedQueueDmTime;
		}else{
			return markerDmTime;
		}
	}
	
	// DirectMusicԂǂł
	void PlayerTimes::ReadMasterClock()
	{
		GuruGuruSmf::Dxmus::Controller::GetInstance()->PortGetReferenceTime(&nowDmTime);
	}
	
	// OɉMĂEFCgāAĂȂ΍̎ԂZbg
	void PlayerTimes::SetMarkerDmTimeByQueueDmTime(__int64 wait)
	{
		__int64 qt = fixedQueueDmTime + wait;
		markerDmTime = qt > nowDmTime ? qt : nowDmTime;
	}
	
	// [vʒuֈړ
	void PlayerTimes::GotoLoopPoint()
	{
		markerTick = loopTick;
		markerTime = loopTime;
	}
		
	// XV
	void PlayerTimes::UpdateMarkerTick(int delta)
	{
		queueDmTime = markerDmTime;
		fixedQueueDmTime = queueDmTime;
		queueTime = markerTime;
		queueTick = markerTick;
		
		markerTick += delta;
		double a = delta * tempo / timeBase;
		markerTime += a / 1000;
		
		// XLbv̓}[J[^CZȂ
		if (!skipping){
			markerDmTime += (__int64)(a*10 / masterTempo);
		}
	}

	// MԂXV
	void PlayerTimes::UpdateQueueDmTime()
	{
		queueDmTime = markerDmTime;
		fixedQueueDmTime = queueDmTime;
	}
	
	// ݂̎ԂZo
	void PlayerTimes::CalcNowTime()
	{
		double a = (double)((nowDmTime + Constant::PrepareTime) - queueDmTime);
		a *= masterTempo;
		nowTime = queueTime + a / 10000;
		nowTick = queueTick + (int)(a * timeBase / (tempo * 10));
	}
	
	// ꎞ~
	void PlayerTimes::Restart()
	{
		__int64 t = nowDmTime - pausedDmTime;
		markerDmTime += t;
		queueDmTime += t;
		fixedQueueDmTime = markerDmTime;
	}
	
	// SMFŏ
	void PlayerTimes::Initialize(Smf* smf)
	{
		timeBase = smf->GetTimeBase();
		
		loopTick = smf->GetData()->LoopTick;
		loopTime = smf->GetData()->LoopTime;
		firstNoteTick = smf->GetData()->FirstNoteTick;
		firstNoteTime = smf->GetData()->FirstNoteTime;
		lastNoteTick = smf->GetData()->LastNoteTick;
		lastNoteTime = smf->GetData()->LastNoteTime;
		
		skipping = false;
	}
	
	// `Ɋ߂
	void PlayerTimes::Rewind()
	{
		SetTempo(Constant::DefaultTempo);
		markerTick = 0;
		markerTime = 0;
		
		queueTime = 0;
		queueTick = 0;
	}
	
	
}

}}



